# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "13/05/2022"


import os
from silx.image.phantomgenerator import PhantomGenerator
from scipy.ndimage import shift as scipy_shift
import numpy
import pytest
from nabu.stitching.config import (
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
)
from nabu.utils import Progress
from nabu.stitching.config import KEY_IMG_REG_METHOD, KEY_SCORE_METHOD, KEY_WINDOW_SIZE
from nabu.stitching.overlap import ZStichOverlapKernel, OverlapStitchingStrategy
from nabu.stitching.z_stitching import (
    PostProcessZStitcher,
    PreProcessZStitcher,
    stitch_vertically_raw_frames,
    ZStitcher,
)
from nxtomomill.nexus.nxtomo import NXtomo
from tomoscan.factory import Factory as TomoscanFactory
from tomoscan.esrf.hdf5scan import ImageKey
from tomoscan.utils.volume import concatenate as concatenate_volumes
from tomoscan.esrf.hdf5scan import HDF5TomoScan
from tomoscan.esrf.volume import HDF5Volume, EDFVolume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume, has_minimal_openjpeg
from tomoscan.esrf.volume.tiffvolume import TIFFVolume, has_tifffile
from nabu.stitching.utils import ShiftAlgorithm
import h5py


strategies_to_test_weights = (
    OverlapStitchingStrategy.CLOSEST,
    OverlapStitchingStrategy.COSINUS_WEIGHTS,
    OverlapStitchingStrategy.LINEAR_WEIGHTS,
    OverlapStitchingStrategy.MEAN,
)


@pytest.mark.parametrize("strategy", strategies_to_test_weights)
def test_overlap_z_stitcher(strategy):
    frame_width = 128
    frame_height = frame_width
    frame_1 = PhantomGenerator.get2DPhantomSheppLogan(n=frame_width)
    stitcher = ZStichOverlapKernel(
        stitching_strategy=strategy,
        overlap_size=frame_height,
        frame_width=128,
    )
    stitched_frame = stitcher.stitch(frame_1, frame_1)[0]
    assert stitched_frame.shape == (frame_height, frame_width)
    # check result is close to the expected one
    numpy.testing.assert_allclose(frame_1, stitched_frame, atol=10e-10)

    # check sum of weights ~ 1.0
    numpy.testing.assert_allclose(
        stitcher.weights_img_1 + stitcher.weights_img_2,
        numpy.ones_like(stitcher.weights_img_1),
    )


@pytest.mark.parametrize("dtype", (numpy.float16, numpy.float32))
def test_z_stitch_raw_frames(dtype):
    """
    test z_stitch_raw_frames: insure a stitching with 3 frames and different overlap can be done
    """
    ref_frame_width = 256
    frame_ref = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(dtype)

    # split the frame into several part
    frame_1 = frame_ref[0:100]
    frame_2 = frame_ref[80:164]
    frame_3 = frame_ref[154:]

    kernel_1 = ZStichOverlapKernel(frame_width=ref_frame_width, overlap_size=20)
    kernel_2 = ZStichOverlapKernel(frame_width=ref_frame_width, overlap_size=10)

    stitched = stitch_vertically_raw_frames(
        frames=(frame_1, frame_2, frame_3),
        output_dtype=dtype,
        overlap_kernels=(kernel_1, kernel_2),
        raw_frames_compositions=None,
        overlap_frames_compositions=None,
        key_lines=(
            (
                90,  # frame_1 height - kernel_1 height / 2.0
                10,  # kernel_1 height / 2.0
            ),
            (
                79,  # frame_2 height - kernel_2 height / 2.0 ou 102-20 ?
                5,  # kernel_2 height / 2.0
            ),
        ),
    )

    assert stitched.shape == frame_ref.shape
    numpy.testing.assert_array_almost_equal(frame_ref, stitched)


def test_z_stitch_raw_frames_2():
    """
    test z_stitch_raw_frames: insure a stitching with 3 frames and different overlap can be done
    """
    ref_frame_width = 256
    frame_ref = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(numpy.float32)

    # split the frame into several part
    frame_1 = frame_ref.copy()
    frame_2 = frame_ref.copy()
    frame_3 = frame_ref.copy()

    kernel_1 = ZStichOverlapKernel(frame_width=ref_frame_width, overlap_size=10)
    kernel_2 = ZStichOverlapKernel(frame_width=ref_frame_width, overlap_size=10)

    stitched = stitch_vertically_raw_frames(
        frames=(frame_1, frame_2, frame_3),
        output_dtype=numpy.float32,
        overlap_kernels=(kernel_1, kernel_2),
        raw_frames_compositions=None,
        overlap_frames_compositions=None,
        key_lines=((20, 20), (105, 105)),
    )

    assert stitched.shape == frame_ref.shape
    numpy.testing.assert_array_almost_equal(frame_ref, stitched)


_stitching_configurations = (
    # simple case where shifts are provided
    {
        "n_proj": 4,
        "raw_pos": ((0, 0, 0), (-90, 0, 0), (-180, 0, 0)),  # requested shift to
        "input_pos": ((0, 0, 0), (-90, 0, 0), (-180, 0, 0)),  # requested shift to
        "raw_shifts": ((0, 0), (-90, 0), (-180, 0)),
    },
    # simple case where shift is found from z position
    {
        "n_proj": 4,
        "raw_pos": ((90, 0, 0), (0, 0, 0), (-90, 0, 0)),
        "input_pos": ((90, 0, 0), (0, 0, 0), (-90, 0, 0)),
        "check_bb": ((40, 140), (-50, 50), (-140, -40)),
        "axis_0_params": {
            KEY_IMG_REG_METHOD: ShiftAlgorithm.NONE,
        },
        "axis_2_params": {
            KEY_IMG_REG_METHOD: ShiftAlgorithm.NONE,
        },
        "raw_shifts": ((0, 0), (-90, 0), (-180, 0)),
    },
)


@pytest.mark.parametrize("configuration", _stitching_configurations)
@pytest.mark.parametrize("dtype", (numpy.float32, numpy.int16))
def test_PreProcessZStitcher(tmp_path, dtype, configuration):
    """
    test PreProcessZStitcher class and insure a full stitching can be done automatically.
    """
    n_proj = configuration["n_proj"]
    ref_frame_width = 280
    raw_frame_height = 100
    ref_frame = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(dtype) * 256.0

    # add some mark for image registration
    ref_frame[:, 96] = -3.2
    ref_frame[:, 125] = 9.1
    ref_frame[:, 165] = 4.4
    ref_frame[:, 200] = -2.5
    # create raw data
    frame_0_shift, frame_1_shift, frame_2_shift = configuration["raw_shifts"]
    frame_0 = scipy_shift(ref_frame, shift=frame_0_shift)[:raw_frame_height]
    frame_1 = scipy_shift(ref_frame, shift=frame_1_shift)[:raw_frame_height]
    frame_2 = scipy_shift(ref_frame, shift=frame_2_shift)[:raw_frame_height]

    frames = frame_0, frame_1, frame_2
    frame_0_input_pos, frame_1_input_pos, frame_2_input_pos = configuration["input_pos"]
    frame_0_raw_pos, frame_1_raw_pos, frame_2_raw_pos = configuration["raw_pos"]

    # create a Nxtomo for each of those raw data
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()
    z_position = (
        frame_0_raw_pos[0],
        frame_1_raw_pos[0],
        frame_2_raw_pos[0],
    )
    scans = []
    for (i_frame, frame), z_pos in zip(enumerate(frames), z_position):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_proj
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_proj, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = numpy.asarray([frame] * n_proj)

        file_path = os.path.join(raw_data_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(HDF5TomoScan(scan=file_path, entry=entry))

    # if requested: check bounding box
    check_bb = configuration.get("check_bb", None)
    if check_bb is not None:
        for scan, expected_bb in zip(scans, check_bb):
            assert scan.get_bounding_box(axis="z") == expected_bb
    output_file_path = os.path.join(output_dir, "stitched.nx")
    output_data_path = "stitched"
    z_stich_config = PreProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        axis_0_pos_px=(
            frame_0_input_pos[0],
            frame_1_input_pos[0],
            frame_2_input_pos[0],
        ),
        axis_1_pos_px=(
            frame_0_input_pos[1],
            frame_1_input_pos[1],
            frame_2_input_pos[1],
        ),
        axis_2_pos_px=(
            frame_0_input_pos[2],
            frame_1_input_pos[2],
            frame_2_input_pos[2],
        ),
        axis_0_pos_mm=None,
        axis_1_pos_mm=None,
        axis_2_pos_mm=None,
        input_scans=scans,
        output_file_path=output_file_path,
        output_data_path=output_data_path,
        axis_0_params=configuration.get("axis_0_params", {}),
        axis_1_params=configuration.get("axis_1_params", {}),
        axis_2_params=configuration.get("axis_2_params", {}),
        output_nexus_version=None,
        slices=None,
        slurm_config=None,
        slice_for_cross_correlation="middle",
        pixel_size=None,
    )
    stitcher = PreProcessZStitcher(z_stich_config)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_file_path
    assert output_identifier.data_path == output_data_path

    created_nx_tomo = NXtomo().load(
        file_path=output_identifier.file_path,
        data_path=output_identifier.data_path,
        detector_data_as="as_numpy_array",
    )

    assert created_nx_tomo.instrument.detector.data.ndim == 3
    mean_abs_error = configuration.get("mean_abs_error", None)
    if mean_abs_error is not None:
        assert (
            numpy.mean(numpy.abs(ref_frame - created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :]))
            < mean_abs_error
        )
    else:
        numpy.testing.assert_array_almost_equal(
            ref_frame, created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :]
        )

    # check also other metadata are here
    assert created_nx_tomo.instrument.detector.distance.value == 2.3
    assert created_nx_tomo.energy.value == 19.2
    numpy.testing.assert_array_equal(
        created_nx_tomo.instrument.detector.image_key_control,
        numpy.asarray([ImageKey.PROJECTION.PROJECTION] * n_proj),
    )

    # check configuration has been saved
    with h5py.File(output_identifier.file_path, mode="r") as h5f:
        assert "stitching_configuration" in h5f[output_identifier.data_path]


slices_to_test_pre = (
    {
        "slices": (None,),
        "complete": True,
    },
    {
        "slices": (("first",), ("middle",), ("last",)),
        "complete": False,
    },
    {
        "slices": ((0, 1, 2), slice(3, -1, 1)),
        "complete": True,
    },
)


@pytest.mark.parametrize("configuration_dist", slices_to_test_pre)
def test_DistributePreProcessZStitcher(tmp_path, configuration_dist):
    slices = configuration_dist["slices"]
    complete = configuration_dist["complete"]

    n_projs = 100
    raw_data = numpy.arange(100 * 128 * 128).reshape((100, 128, 128))

    # create raw data
    frame_0 = raw_data[:, 60:]
    assert frame_0.ndim == 3
    frame_0_pos = 40
    frame_1 = raw_data[:, 0:80]
    assert frame_1.ndim == 3
    frame_1_pos = 94
    frames = (frame_0, frame_1)
    z_positions = (frame_0_pos, frame_1_pos)

    # create a Nxtomo for each of those raw data
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    scans = []
    for (i_frame, frame), z_pos in zip(enumerate(frames), z_positions):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_projs
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_projs, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_projs
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = frame

        file_path = os.path.join(raw_data_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(HDF5TomoScan(scan=file_path, entry=entry))

    stitched_nx_tomo = []
    for s in slices:
        output_file_path = os.path.join(output_dir, "stitched_section.nx")
        output_data_path = f"stitched_{s}"
        z_stich_config = PreProcessedZStitchingConfiguration(
            axis_0_pos_px=z_positions,
            axis_1_pos_px=None,
            axis_2_pos_px=(0, 0),
            axis_0_pos_mm=None,
            axis_1_pos_mm=None,
            axis_2_pos_mm=None,
            axis_0_params={},
            axis_1_params={},
            axis_2_params={},
            stitching_strategy=OverlapStitchingStrategy.CLOSEST,
            overwrite_results=True,
            input_scans=scans,
            output_file_path=output_file_path,
            output_data_path=output_data_path,
            output_nexus_version=None,
            slices=s,
            slurm_config=None,
            slice_for_cross_correlation="middle",
            pixel_size=None,
        )
        stitcher = PreProcessZStitcher(z_stich_config)
        output_identifier = stitcher.stitch()
        assert output_identifier.file_path == output_file_path
        assert output_identifier.data_path == output_data_path

        created_nx_tomo = NXtomo().load(
            file_path=output_identifier.file_path,
            data_path=output_identifier.data_path,
            detector_data_as="as_numpy_array",
        )
        stitched_nx_tomo.append(created_nx_tomo)
    assert len(stitched_nx_tomo) == len(slices)
    final_nx_tomo = NXtomo.concatenate(stitched_nx_tomo)
    assert isinstance(final_nx_tomo.instrument.detector.data, numpy.ndarray)
    final_nx_tomo.save(
        file_path=os.path.join(output_dir, "final_stitched.nx"),
        data_path="entry0000",
    )

    if complete:
        len(final_nx_tomo.instrument.detector.data) == 128
        # test middle
        numpy.testing.assert_array_almost_equal(raw_data[1], final_nx_tomo.instrument.detector.data[1, :, :])
    else:
        len(final_nx_tomo.instrument.detector.data) == 3
        # test middle
        numpy.testing.assert_array_almost_equal(raw_data[49], final_nx_tomo.instrument.detector.data[1, :, :])
    # in the case of first, middle and last frames
    # test first
    numpy.testing.assert_array_almost_equal(raw_data[0], final_nx_tomo.instrument.detector.data[0, :, :])

    # test last
    numpy.testing.assert_array_almost_equal(raw_data[-1], final_nx_tomo.instrument.detector.data[-1, :, :])


_VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING = [HDF5Volume, EDFVolume]
# avoid testing glymur because doesn't handle float
# if has_minimal_openjpeg:
#     _VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING.append(JP2KVolume)
if has_tifffile:
    _VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING.append(TIFFVolume)


@pytest.mark.parametrize("progress", (None, Progress("z-stitching")))
@pytest.mark.parametrize("volume_class", (_VOL_CLASSES_TO_TEST_FOR_POSTPROC_STITCHING))
def test_PostProcessZStitcher(
    tmp_path,
    volume_class,
    progress,
):
    """
    test PreProcessZStitcher class and insure a full stitching can be done automatically.

    :param bool clear_input_volumes_data: if True save the volume then clear volume.data (used to check internal management of loading volumes - used to check behavior with HDF5)
    :param volume_class: class to be used (same class for input and output for now)
    :param axis_0_pos: position of the different TomoObj along axis 0 (Also know as z axis)
    """
    # create some random data.
    raw_volume = numpy.stack(
        [
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 256.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 128.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 32.0,
            PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float32) * 16.0,
        ]
    )
    assert raw_volume.shape == (4, 120, 120)
    raw_volume = numpy.rollaxis(raw_volume, axis=1, start=0)
    assert raw_volume.shape == (120, 4, 120)

    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    # create a simple case where the volume have 10 voxel of overlap and a height (z) of 30 Voxels, 40 and 30 Voxels
    vol_1_constructor_params = {
        "data": raw_volume[0:30, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-15.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_2_constructor_params = {
        "data": raw_volume[20:80, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    vol_3_constructor_params = {
        "data": raw_volume[60:, :, :],
        "metadata": {
            "processing_options": {
                "reconstruction": {
                    "position": (-90.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    }

    raw_volumes = []
    axis_0_positions = []
    for i_vol, vol_params in enumerate([vol_1_constructor_params, vol_2_constructor_params, vol_3_constructor_params]):
        if volume_class == HDF5Volume:
            vol_params.update(
                {
                    "file_path": os.path.join(raw_data_dir, f"raw_volume_{i_vol}.hdf5"),
                    "data_path": "volume",
                }
            )
        else:
            vol_params.update(
                {
                    "folder": os.path.join(raw_data_dir, f"raw_volume_{i_vol}"),
                }
            )
        axis_0_positions.append(vol_params["metadata"]["processing_options"]["reconstruction"]["position"][0])

        volume = volume_class(**vol_params)
        volume.save()
        raw_volumes.append(volume)

    volume_1, volume_2, volume_3 = raw_volumes

    output_volume = HDF5Volume(
        file_path=os.path.join(output_dir, "stitched_volume.hdf5"),
        data_path="stitched_volume",
    )

    z_stich_config = PostProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_volumes=(volume_1, volume_2, volume_3),
        output_volume=output_volume,
        slices=None,
        slurm_config=None,
        axis_0_pos_px=axis_0_positions,
        axis_0_pos_mm=None,
        axis_0_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_1_pos_px=None,
        axis_1_pos_mm=None,
        axis_1_params={"img_reg_method": ShiftAlgorithm.NONE},
        axis_2_pos_px=None,
        axis_2_pos_mm=None,
        axis_2_params={"img_reg_method": ShiftAlgorithm.NONE},
        slice_for_cross_correlation="middle",
        voxel_size=None,
    )

    stitcher = PostProcessZStitcher(z_stich_config, progress=progress)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_volume.file_path
    assert output_identifier.data_path == output_volume.data_path

    output_volume.data = None
    output_volume.metadata = None
    output_volume.load_data(store=True)
    output_volume.load_metadata(store=True)

    assert raw_volume.shape == output_volume.data.shape

    numpy.testing.assert_array_almost_equal(raw_volume, output_volume.data)

    metadata = output_volume.metadata
    assert metadata["program"] == "nabu-stitching"
    assert "configuration" in metadata
    assert output_volume.position[0] == -60.0
    assert output_volume.pixel_size == (1.0, 1.0, 1.0)


slices_to_test_post = (
    {
        "slices": (None,),
        "complete": True,
    },
    {
        "slices": (("first",), ("middle",), ("last",)),
        "complete": False,
    },
    {
        "slices": ((0, 1, 2), slice(3, -1, 1)),
        "complete": True,
    },
)


@pytest.mark.parametrize("flip_ud", (True, False))
@pytest.mark.parametrize("configuration_dist", slices_to_test_post)
def test_DistributePostProcessZStitcher(tmp_path, configuration_dist, flip_ud):
    # create some random data.
    slices = configuration_dist["slices"]
    complete = configuration_dist["complete"]

    raw_volume = numpy.ones((80, 40, 120), dtype=numpy.float16)
    raw_volume[:, 0, :] = (
        PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float16)[30:110, :] * 80 * 40 * 120
    )
    raw_volume[:, 8, :] = (
        PhantomGenerator.get2DPhantomSheppLogan(n=120).astype(numpy.float16)[30:110, :] * 80 * 40 * 120 + 2
    )
    raw_volume[12] = 1.0
    raw_volume[:, 23] = 1.2
    # create folder to save data (and debug)
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    def flip_input_data(data):
        if flip_ud:
            data = numpy.flipud(data)
        return data

    volume_1 = HDF5Volume(
        file_path=os.path.join(raw_data_dir, "volume_1.hdf5"),
        data_path="volume",
        data=flip_input_data(raw_volume[-60:, :, :]),
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-30.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    )
    volume_1.save()

    volume_2 = HDF5Volume(
        file_path=os.path.join(raw_data_dir, "volume_2.hdf5"),
        data_path="volume",
        data=flip_input_data(raw_volume[:60, :, :]),
        metadata={
            "processing_options": {
                "reconstruction": {
                    "position": (-50.0, 0.0, 0.0),
                    "voxel_size_cm": (100.0, 100.0, 100.0),
                }
            },
        },
    )
    volume_2.save()

    reconstructed_sub_volumes = []
    for i_slice, s in enumerate(slices):
        output_volume = HDF5Volume(
            file_path=os.path.join(output_dir, f"stitched_subvolume_{i_slice}.hdf5"),
            data_path="stitched_volume",
        )
        volumes = (volume_2, volume_1)
        z_stich_config = PostProcessedZStitchingConfiguration(
            stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
            axis_0_pos_px=tuple(
                volume.metadata["processing_options"]["reconstruction"]["position"][0] for volume in volumes
            ),
            axis_0_pos_mm=None,
            axis_0_params={},
            axis_1_pos_px=None,
            axis_1_pos_mm=None,
            axis_1_params={},
            axis_2_pos_px=(0, 0),
            axis_2_pos_mm=None,
            axis_2_params={},
            overwrite_results=True,
            input_volumes=volumes,
            output_volume=output_volume,
            slices=s,
            slurm_config=None,
            slice_for_cross_correlation="middle",
            voxel_size=None,
            flip_ud=flip_ud,
        )

        stitcher = PostProcessZStitcher(z_stich_config)
        vol_id = stitcher.stitch()
        reconstructed_sub_volumes.append(TomoscanFactory.create_tomo_object_from_identifier(identifier=vol_id))

    final_vol = HDF5Volume(
        file_path=os.path.join(output_dir, "final_volume"),
        data_path="volume",
    )
    if complete:
        concatenate_volumes(output_volume=final_vol, volumes=tuple(reconstructed_sub_volumes), axis=1)
        final_vol.load_data(store=True)
        with h5py.File("debug.hdf5", mode="w") as h5f:
            h5f["raw_volume"] = raw_volume
            h5f["final_vol"] = final_vol.data
            h5f["volume1"] = volume_1.data
            h5f["volume2"] = volume_2.data
        numpy.testing.assert_almost_equal(
            raw_volume,
            final_vol.data,
        )


def test_get_overlap_areas():
    """test get_overlap_areas function"""
    f_upper = numpy.linspace(7, 15, num=9, endpoint=True)
    f_lower = numpy.linspace(0, 12, num=13, endpoint=True)

    o_1, o_2 = ZStitcher.get_overlap_areas(
        upper_frame=f_upper,
        lower_frame=f_lower,
        upper_frame_key_line=3,
        lower_frame_key_line=10,
        overlap_size=4,
        stitching_axis=0,
    )

    numpy.testing.assert_array_equal(o_1, o_2)
    numpy.testing.assert_array_equal(o_1, numpy.linspace(8, 11, num=4, endpoint=True))


def test_frame_flip(tmp_path):
    """check it with some NXtomo fliped"""
    ref_frame_width = 280
    n_proj = 10
    raw_frame_width = 100
    ref_frame = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(numpy.float32) * 256.0
    # create raw data
    frame_0_shift = (0, 0)
    frame_1_shift = (-90, 0)
    frame_2_shift = (-180, 0)

    frame_0 = scipy_shift(ref_frame, shift=frame_0_shift)[:raw_frame_width]
    frame_1 = scipy_shift(ref_frame, shift=frame_1_shift)[:raw_frame_width]
    frame_2 = scipy_shift(ref_frame, shift=frame_2_shift)[:raw_frame_width]
    frames = frame_0, frame_1, frame_2

    x_flips = [False, True, True]
    y_flips = [False, False, True]

    def apply_flip(args):
        frame, flip_x, flip_y = args
        if flip_x:
            frame = numpy.fliplr(frame)
        if flip_y:
            frame = numpy.flipud(frame)
        return frame

    frames = map(apply_flip, zip(frames, x_flips, y_flips))

    # create a Nxtomo for each of those raw data
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()
    z_position = (90, 0, -90)

    scans = []
    for (i_frame, frame), z_pos, x_flip, y_flip in zip(enumerate(frames), z_position, x_flips, y_flips):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_proj
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_proj, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.instrument.detector.x_flipped = x_flip
        nx_tomo.instrument.detector.y_flipped = y_flip
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = numpy.asarray([frame] * n_proj)

        file_path = os.path.join(raw_data_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(HDF5TomoScan(scan=file_path, entry=entry))

    output_file_path = os.path.join(output_dir, "stitched.nx")
    output_data_path = "stitched"
    assert len(scans) == 3
    z_stich_config = PreProcessedZStitchingConfiguration(
        axis_0_pos_px=(0, -90, -180),
        axis_1_pos_px=None,
        axis_2_pos_px=(0, 0, 0),
        axis_0_pos_mm=None,
        axis_1_pos_mm=None,
        axis_2_pos_mm=None,
        axis_0_params={},
        axis_1_params={},
        axis_2_params={},
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_scans=scans,
        output_file_path=output_file_path,
        output_data_path=output_data_path,
        output_nexus_version=None,
        slices=None,
        slurm_config=None,
        slice_for_cross_correlation="middle",
        pixel_size=None,
    )
    stitcher = PreProcessZStitcher(z_stich_config)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_file_path
    assert output_identifier.data_path == output_data_path

    created_nx_tomo = NXtomo().load(
        file_path=output_identifier.file_path,
        data_path=output_identifier.data_path,
        detector_data_as="as_numpy_array",
    )

    assert created_nx_tomo.instrument.detector.data.ndim == 3
    # insure flipping has been taking into account
    numpy.testing.assert_array_almost_equal(ref_frame, created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :])

    assert created_nx_tomo.instrument.detector.x_flipped is False
    assert created_nx_tomo.instrument.detector.y_flipped is False



from nabu.stitching.overlap import compute_image_minimum_divergence, compute_image_higher_signal
from nabu.testutils import get_data
import numpy


def test_compute_image_minimum_divergence():
    """make sure the compute_image_minimum_divergence function is processing"""
    raw_data_1 = get_data("brain_phantom.npz")["data"]
    raw_data_2 = numpy.random.rand(*raw_data_1.shape) * 255.0

    stitching = compute_image_minimum_divergence(raw_data_1, raw_data_2, high_frequency_threshold=2)
    assert stitching.shape == raw_data_1.shape


def test_compute_image_higher_signal():
    """
    make sure compute_image_higher_signal is processing
    """
    raw_data = get_data("brain_phantom.npz")["data"]
    raw_data_1 = raw_data.copy()
    raw_data_1[40:75] = 0.0
    raw_data_1[:, 210:245] = 0.0

    raw_data_2 = raw_data.copy()
    raw_data_2[:, 100:120] = 0.0

    stitching = compute_image_higher_signal(raw_data_1, raw_data_2)

    numpy.testing.assert_array_equal(
        stitching,
        raw_data,
    )



from copy import copy
from dataclasses import dataclass
import numpy
from math import ceil

from nabu.stitching.overlap import ZStichOverlapKernel


@dataclass
class _FrameCompositionBase:
    def compose(self, output_frame: numpy.ndarray, input_frames: tuple):
        raise NotImplementedError("Base class")


@dataclass
class ZFrameComposition(_FrameCompositionBase):
    """
    class used to define intervals to know where to dump raw data or stitched data according to requested policy.
    The idea is to create this once for all for one stitching operation and reuse it for each frame.
    """

    local_start_y: tuple
    local_end_y: tuple
    global_start_y: tuple
    global_end_y: tuple

    def browse(self):
        for i in range(len(self.local_start_y)):
            yield (
                self.local_start_y[i],
                self.local_end_y[i],
                self.global_start_y[i],
                self.global_end_y[i],
            )

    def compose(self, output_frame: numpy.ndarray, input_frames: tuple):
        if not output_frame.ndim == 2:
            raise TypeError(f"output_frame is expected to be 2D and not {output_frame.ndim}")
        for (
            global_start_y,
            global_end_y,
            local_start_y,
            local_end_y,
            input_frame,
        ) in zip(
            self.global_start_y,
            self.global_end_y,
            self.local_start_y,
            self.local_end_y,
            input_frames,
        ):
            if input_frame is not None:
                output_frame[global_start_y:global_end_y] = input_frame[local_start_y:local_end_y]

    @staticmethod
    def compute_raw_frame_compositions(frames: tuple, key_lines: tuple, overlap_kernels: tuple, stitching_axis):
        """
        compute frame composition for raw data

        warning: we expect frames to be ordered y downward and the frame order to keep this ordering
        """
        assert len(frames) == len(overlap_kernels) + 1 == len(key_lines) + 1

        global_start_ys = [0]

        # extend shifts and kernels to have a first shift of 0 and two overlaps values at 0 to
        # generalize processing
        local_start_ys = [0]

        local_start_ys.extend(
            [ceil(key_line[1] + kernel.overlap_size / 2) for (key_line, kernel) in zip(key_lines, overlap_kernels)]
        )
        local_end_ys = list(
            [ceil(key_line[0] - kernel.overlap_size / 2) for (key_line, kernel) in zip(key_lines, overlap_kernels)]
        )
        local_end_ys.append(frames[-1].shape[stitching_axis])

        for (
            new_local_start_y,
            new_local_end_y,
            kernel,
        ) in zip(local_start_ys, local_end_ys, overlap_kernels):
            global_start_ys.append(global_start_ys[-1] + (new_local_end_y - new_local_start_y) + kernel.overlap_size)

        # global end can be easily found from global start + local start and end
        global_end_ys = []
        for global_start_y, new_local_start_y, new_local_end_y in zip(global_start_ys, local_start_ys, local_end_ys):
            global_end_ys.append(global_start_y + new_local_end_y - new_local_start_y)

        return ZFrameComposition(
            local_start_y=tuple(local_start_ys),
            local_end_y=tuple(local_end_ys),
            global_start_y=tuple(global_start_ys),
            global_end_y=tuple(global_end_ys),
        )

    @staticmethod
    def compute_stitch_frame_composition(frames, key_lines: tuple, overlap_kernels: tuple, stitching_axis: int):
        """
        compute frame composition for stiching.
        """
        assert len(frames) == len(overlap_kernels) + 1 == len(key_lines) + 1
        assert stitching_axis in (0, 1, 2)

        # position in the stitched frame;
        local_start_ys = [0] * len(overlap_kernels)
        local_end_ys = [kernel.overlap_size for kernel in overlap_kernels]

        # position in the global frame. For this one it is simpler to rely on the raw frame composition
        composition_raw = ZFrameComposition.compute_raw_frame_compositions(
            frames=frames,
            key_lines=key_lines,
            overlap_kernels=overlap_kernels,
            stitching_axis=stitching_axis,
        )
        global_start_ys = composition_raw.global_end_y[:-1]
        global_end_ys = composition_raw.global_start_y[1:]

        return ZFrameComposition(
            local_start_y=tuple(local_start_ys),
            local_end_y=tuple(local_end_ys),
            global_start_y=tuple(global_start_ys),
            global_end_y=tuple(global_end_ys),
        )

    @staticmethod
    def pprint_z_stitching(raw_composition, stitch_composition):
        """
        util to display what the output of the z stitch will looks like from composition
        """
        for i_frame, (raw_comp, stitch_comp) in enumerate(zip(raw_composition.browse(), stitch_composition.browse())):
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = raw_comp

            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )

            (
                stitch_local_start,
                stitch_local_end,
                stitch_global_start,
                stitch_global_end,
            ) = stitch_comp

            print(
                f"stitch_frame[{stitch_global_start}:{stitch_global_end}] = stitched_frame_{i_frame}[{stitch_local_start}:{stitch_local_end}]"
            )
        else:
            i_frame += 1
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = list(raw_composition.browse())[-1]
            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )



import pytest
import numpy as np
from nabu.testutils import utilstest
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from nabu.pipeline.estimators import CompositeCOREstimator


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    dataset_downloaded_path = utilstest.getfile("test_composite_cor_finder_data.h5")
    cls.theta_interval = 4.5 * 1  # this is given. Radios in the middle of steps 4.5 degree long
    # are set to zero for compression
    # You can still change it to a multiple of 4.5
    cls.cor_pix = 1321.625
    cls.abs_tol = 0.0001
    cls.dataset_info = HDF5DatasetAnalyzer(dataset_downloaded_path)
    cls.cor_options = """side="near"; near_pos = 300.0;  near_width = 20.0 """


@pytest.mark.usefixtures("bootstrap")
class TestCompositeCorFinder:
    def test(self):
        cor_finder = CompositeCOREstimator(
            self.dataset_info, theta_interval=self.theta_interval, cor_options=self.cor_options
        )

        cor_position = cor_finder.find_cor()
        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_pix
        assert np.isclose(self.cor_pix, cor_position, atol=self.abs_tol), message



import numpy as np
from ..thirdparty.pore3d_deringer_munch import munchetal_filter
from ..utils import get_2D_3D_shape


class MunchDeringer:
    def __init__(self, sigma, sinos_shape, levels=None, wname="db15", padding=None, padding_mode="edge"):
        """
        Initialize a "Munch Et Al" sinogram deringer. See References for more information.

        Parameters
        -----------
        sigma: float
            Standard deviation of the damping parameter. The higher value of sigma,
            the more important the filtering effect on the rings.
        levels: int, optional
            Number of wavelets decomposition levels.
            By default (None), the maximum number of decomposition levels is used.
        wname: str, optional
            Default is "db15" (Daubechies, 15 vanishing moments)
        sinos_shape: tuple, optional
            Shape of the sinogram (or sinograms stack).
        padding: tuple of two int, optional
            Horizontal padding to use for reducing the aliasing artefacts

        References
        ----------
        B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
        combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
        """
        self._get_shapes(sinos_shape, padding)
        self.sigma = sigma
        self.levels = levels
        self.wname = wname
        self.padding_mode = padding_mode
        self._check_can_use_wavelets()

    def _get_shapes(self, sinos_shape, padding):
        n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
        self.sinos_shape = n_z, n_a, n_x
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x
        # Handle "padding=True" or "padding=False"
        if isinstance(padding, bool):
            if padding:
                padding = (n_x // 2, n_x // 2)
            else:
                padding = None
        #
        if padding is not None:
            pad_x1, pad_x2 = padding
            if np.iterable(pad_x1) or np.iterable(pad_x2):
                raise ValueError("Expected padding in the form (x1, x2)")
            self.sino_padded_shape = (n_a, n_x + pad_x1 + pad_x2)
        self.padding = padding

    def _check_can_use_wavelets(self):
        if munchetal_filter is None:
            raise ValueError("Need pywavelets to use this class")

    def _destripe_2D(self, sino, output):
        if self.padding is not None:
            sino = np.pad(sino, ((0, 0), self.padding), mode=self.padding_mode)
        res = munchetal_filter(sino, self.levels, self.sigma, wname=self.wname)
        if self.padding is not None:
            res = res[:, self.padding[0] : -self.padding[1]]
        output[:] = res
        return output

    def remove_rings(self, sinos, output=None):
        """
        Main function to performs rings artefacts removal on sinogram(s).
        CAUTION: this function defaults to in-place processing, meaning that
        the sinogram(s) you pass will be overwritten.

        Parameters
        ----------
        sinos: numpy.ndarray
            Sinogram or stack of sinograms.
        output: numpy.ndarray, optional
            Output array. If set to None (default), the output overwrites the input.
        """
        if output is None:
            output = sinos
        if sinos.ndim == 2:
            return self._destripe_2D(sinos, output)
        n_sinos = sinos.shape[0]
        for i in range(n_sinos):
            self._destripe_2D(sinos[i], output[i])
        return output



import numpy as np
import pytest
from nabu.utils import clip_circle
from nabu.testutils import get_data, compare_arrays, generate_tests_scenarios, __do_long_tests__
from nabu.reconstruction.rings import MunchDeringer
from nabu.thirdparty.pore3d_deringer_munch import munchetal_filter
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

__have_gpuderinger__ = False
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.reconstruction.rings_cuda import CudaMunchDeringer, __have_pycudwt__, __have_skcuda__

    if __have_pycudwt__ and __have_skcuda__:
        __have_gpuderinger__ = True


fw_scenarios = generate_tests_scenarios(
    {
        "levels": [4],
        "sigma": [1.0],
        "wname": ["db15"],
        "padding": [None],
    }
)
if __do_long_tests__:
    fw_scenarios = generate_tests_scenarios(
        {
            "levels": [4, 2],
            "sigma": [1.0, 2.0],
            "wname": ["db15", "haar", "rbio4.4"],
            "padding": [None, (100, 100), (50, 71)],
        }
    )


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("mri_sino500.npz")["data"]
    cls.tol = 5e-3
    cls.rings = {150: 0.5, -150: 0.5}
    if __have_gpuderinger__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)
    yield
    if __have_gpuderinger__:
        cls.ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestMunchDeringer:
    @staticmethod
    def add_stripes_to_sino(sino, rings_desc):
        """
        Create a new sinogram by adding synthetic stripes to an existing one.

        Parameters
        ----------
        sino: array-like
            Sinogram.
        rings_desc: dict
            Dictionary describing the stripes locations and intensity.
            The location is an integer in [0, N[ where N is the number of columns.
            The intensity is a float: percentage of the current column mean value.
        """
        sino_out = np.copy(sino)
        for loc, intensity in rings_desc.items():
            sino_out[:, loc] += sino[:, loc].mean() * intensity
        return sino_out

    @staticmethod
    def get_fourier_wavelets_reference_result(sino, config):
        # Reference destriping with pore3d "munchetal_filter"
        padding = config.get("padding", None)
        if padding is not None:
            sino = np.pad(sino, ((0, 0), padding), mode="edge")
        ref = munchetal_filter(sino, config["levels"], config["sigma"], wname=config["wname"])
        if config["padding"] is not None:
            ref = ref[:, padding[0] : -padding[1]]
        return ref

    @pytest.mark.skipif(munchetal_filter is None, reason="Need PyWavelets for this test")
    @pytest.mark.parametrize("config", fw_scenarios)
    def test_munch_deringer(self, config):
        deringer = MunchDeringer(
            config["sigma"], self.sino.shape, levels=config["levels"], wname=config["wname"], padding=config["padding"]
        )
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        ref = self.get_fourier_wavelets_reference_result(sino, config)
        # Wrapping with DeRinger
        res = np.zeros((1,) + sino.shape, dtype=np.float32)
        deringer.remove_rings(sino, output=res)

        err_max = np.max(np.abs(res[0] - ref))
        assert err_max < self.tol, "Max error is too high"

    @pytest.mark.skipif(
        not (__have_gpuderinger__) or munchetal_filter is None,
        reason="Need pycuda, pycudwt and scikit-cuda for this test",
    )
    @pytest.mark.parametrize("config", fw_scenarios)
    def test_cuda_munch_deringer(self, config):
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        deringer = CudaMunchDeringer(
            config["sigma"],
            self.sino.shape,
            levels=config["levels"],
            wname=config["wname"],
            cuda_options={"ctx": self.ctx},
        )
        d_sino = garray.to_gpu(sino)
        deringer.remove_rings(d_sino)
        res = d_sino.get()

        ref = self.get_fourier_wavelets_reference_result(sino, config)

        err_max = np.max(np.abs(res - ref))
        assert err_max < 1e-1, "Max error is too high"



import pytest
from tomoscan.test.utils import HDF5MockContext


@pytest.fixture(scope="function")
def hdf5_scan(tmp_path):
    """simple fixture to create a scan and provide it to another function"""
    test_dir = tmp_path / "my_hdf5_scan"
    with HDF5MockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        yield scan



import numpy as np


class ProcessingBase:
    _array_class = None

    def __init__(self):
        self._allocated = {}

    def init_arrays_to_none(self, arrays_names):
        """
        Initialize arrays to None. After calling this method, the current instance will
        have self.array_name = None, and self._old_array_name = None.

        Parameters
        ----------
        arrays_names: list of str
            List of arrays names.
        """
        for array_name in arrays_names:
            setattr(self, array_name, None)
            setattr(self, "_old_" + array_name, None)
            self._allocated[array_name] = False

    def recover_arrays_references(self, arrays_names):
        """
        Performs self._array_name = self._old_array_name,
        for each array_name in arrays_names.

        Parameters
        ----------
        arrays_names: list of str
            List of array names
        """
        for array_name in arrays_names:
            old_arr = getattr(self, "_old_" + array_name, None)
            if old_arr is not None:
                setattr(self, array_name, old_arr)

    def _allocate_array_mem(self, shape, dtype):
        raise ValueError("Base class")

    def allocate_array(self, array_name, shape, dtype=np.float32):
        """
        Allocate a GPU array on the current context/stream/device,
        and set 'self.array_name' to this array.

        Parameters
        ----------
        array_name: str
            Name of the array (for book-keeping)
        shape: tuple of int
            Array shape
        dtype: numpy.dtype, optional
            Data type. Default is float32.
        """
        if not self._allocated.get(array_name, False):
            new_device_arr = self._allocate_array_mem(shape, dtype)
            setattr(self, array_name, new_device_arr)
            self._allocated[array_name] = True
        return getattr(self, array_name)

    def set_array(self, array_name, array_ref, dtype=np.float32):
        """
        Set the content of a device array.

        Parameters
        ----------
        array_name: str
            Array name. This method will look for self.array_name.
        array_ref: array (numpy or GPU array)
            Array containing the data to copy to 'array_name'.
        dtype: numpy.dtype, optional
            Data type. Default is float32.
        """
        if isinstance(array_ref, self._array_class):
            current_arr = getattr(self, array_name, None)
            setattr(self, "_old_" + array_name, current_arr)
            setattr(self, array_name, array_ref)
        elif isinstance(array_ref, np.ndarray):
            self.allocate_array(array_name, array_ref.shape, dtype=dtype)
            getattr(self, array_name).set(array_ref)
        else:
            raise ValueError("Expected numpy array or pycuda array")
        return getattr(self, array_name)

    def get_array(self, array_name):
        return getattr(self, array_name, None)

    # COMPAT.
    _init_arrays_to_none = init_arrays_to_none
    _recover_arrays_references = recover_arrays_references
    _allocate_array = allocate_array
    _set_array = set_array

    def check_array(self, arr, expected_shape, expected_dtype="f", check_contiguous=True):
        """
        Check whether a given array is suitable for being processed (shape, dtype, contiguous)
        """
        if arr.shape != expected_shape:
            raise ValueError("Expected shape %s but got %s" % (str(expected_shape), str(arr.shape)))
        if arr.dtype != np.dtype(expected_dtype):
            raise ValueError("Expected data type %s but got %s" % (str(expected_dtype), str(arr.dtype)))
        if check_contiguous:
            if isinstance(arr, np.ndarray) and not (arr.flags["C_CONTIGUOUS"]):
                raise ValueError("Expected C-contiguous array")
            if isinstance(arr, self._array_class) and not arr.flags.c_contiguous:
                raise ValueError("Expected C-contiguous array")



from nabu.stitching.utils import shift_grid_search, ScoreMethod, has_itk, find_shift_with_itk
import scipy.misc
from scipy.ndimage import shift as scipy_shift
from scipy.ndimage import shift as shift_scipy
import numpy
import pytest


@pytest.mark.parametrize("shift", [(0, 3), (-4, 0), (8, -6)])
def test_shift_grid_search(shift):
    """
    test shift_grid_search algorithm
    """
    y = numpy.sin(numpy.linspace(0, numpy.pi, 250))

    weights = numpy.cos(numpy.linspace(-numpy.pi, numpy.pi, 120))
    image = numpy.outer(y, weights)
    # add a simple line with different value to ease detection
    image[150] = 1.0
    image[:, 50] = 1.2

    image_ref = image
    image_with_shift = scipy_shift(image_ref.copy(), shift=-numpy.array(shift))
    score_method = ScoreMethod.TV

    best_shift = shift_grid_search(
        image_ref,
        image_with_shift,
        window_sizes=(40, 20),
        axis=(0, 1),
        step_size=1,
        score_method=score_method,
    )

    assert tuple(best_shift) == shift


@pytest.mark.parametrize("data_type", (numpy.float32, numpy.uint16))
@pytest.mark.skipif(not has_itk, reason="itk not installed")
def test_find_shift_with_itk(data_type):
    shift = (5, 2)
    img1 = scipy.misc.ascent().astype(data_type)
    img2 = shift_scipy(
        img1.copy(),
        shift=shift,
        order=1,
    )

    img1 = img1[10:-10, 10:-10]
    img2 = img2[10:-10, 10:-10]
    assert find_shift_with_itk(img1=img1, img2=img2) == shift



import os
import copy
import numpy
from silx.io.url import DataUrl
from typing import Iterable, Optional, Union
from tomoscan.tomoobject import TomoObject
from nabu.app.bootstrap_stitching import _SECTIONS_COMMENTS
from nabu.stitching.config import (
    StitchingConfiguration,
    get_default_stitching_config,
    PreProcessedZStitchingConfiguration,
    PostProcessedZStitchingConfiguration,
    SLURM_SECTION,
)
from nabu.pipeline.config import generate_nabu_configfile
from tomoscan.esrf import HDF5TomoScan
from tomoscan.esrf import EDFTomoScan
from tomoscan.esrf.volume import HDF5Volume, MultiTIFFVolume
from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase

try:
    from sluurp.job import SBatchScriptJob
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True


def split_stitching_configuration_to_slurm_job(
    configuration: StitchingConfiguration, yield_configuration: bool = False
):
    """
    generator to split a StitchingConfiguration into several SBatchScriptJob.

    This will handle:
       * division into several jobs according to `slices` and `n_job`
       * creation of SBatchScriptJob handling slurm configuration and command to be launched

    :param StitchingConfiguration configuration: configuration of the stitching to launch (into several jobs)
    :param bool yield_configuration: if True then yield (SBatchScriptJob, StitchingConfiguration) else yield only SBatchScriptJob
    """
    if not isinstance(configuration, StitchingConfiguration):
        raise TypeError(
            f"configuration is expected to be an instance of {StitchingConfiguration}. {type(configuration)} provided."
        )
    if not has_sluurp:
        raise ImportError("sluurp not install. Please install it to distribute stitching on slurm (pip install sluurm)")

    slurm_configuration = configuration.slurm_config
    n_jobs = slurm_configuration.n_jobs
    stitching_type = configuration.stitching_type
    # cleqn slurm configurqtion
    slurm_configuration = slurm_configuration.to_dict()
    clean_script = slurm_configuration.pop("clean_scripts", False)
    slurm_configuration.pop("n_jobs", None)
    # for now other_options is not handled
    slurm_configuration.pop("other_options", None)

    if "memory" in slurm_configuration and isinstance(slurm_configuration["memory"], str):
        memory = slurm_configuration["memory"].lower().replace(" ", "")
        memory = memory.rstrip("b").rstrip("g")
        slurm_configuration["memory"] = memory
    # handle slices if None
    configuration.settle_inputs()

    slice_sub_parts = split_slices(slices=configuration.slices, n_parts=n_jobs)
    if isinstance(configuration, PreProcessedZStitchingConfiguration):
        stitch_prefix = os.path.basename(os.path.splitext(configuration.output_file_path)[0])
        configuration.output_file_path = os.path.abspath(configuration.output_file_path)

    elif isinstance(configuration, PostProcessedZStitchingConfiguration):
        stitch_prefix = os.path.basename(os.path.splitext(configuration.output_volume.file_path)[0])
        configuration.output_volume.file_path = os.path.abspath(configuration.output_volume.file_path)
    else:
        raise TypeError(f"{type(configuration)} not handled")

    for i_sub_part, slice_sub_part in enumerate(slice_sub_parts):
        sub_configuration = copy.deepcopy(configuration)
        # update slice
        sub_configuration.slices = slice_sub_part
        # remove slurm configuration because once on the partition we run it manually
        sub_configuration.slurm_config = None

        if isinstance(sub_configuration, PreProcessedZStitchingConfiguration):
            original_output_file_path, file_extension = os.path.splitext(sub_configuration.output_file_path)
            sub_configuration.output_file_path = os.path.join(
                original_output_file_path,
                os.path.basename(original_output_file_path) + f"_part_{i_sub_part}" + file_extension,
            )
            output_obj = HDF5TomoScan(
                scan=sub_configuration.output_file_path,
                entry=sub_configuration.output_data_path,
            )
        elif isinstance(sub_configuration, PostProcessedZStitchingConfiguration):
            if isinstance(sub_configuration.output_volume, (HDF5Volume, MultiTIFFVolume)):
                original_output_file_path, file_extension = os.path.splitext(sub_configuration.output_volume.file_path)
                sub_configuration.output_volume.file_path = os.path.join(
                    original_output_file_path,
                    os.path.basename(original_output_file_path) + f"_part_{i_sub_part}" + file_extension,
                )
            elif isinstance(sub_configuration.output_volume, VolumeSingleFrameBase):
                url = sub_configuration.output_volume.url
                original_output_folder = url.file_path()
                sub_part_url = DataUrl(
                    file_path=os.path.join(
                        original_output_folder,
                        os.path.basename(original_output_folder) + f"_part_{i_sub_part}",
                    ),
                    data_path=url.data_path(),
                    scheme=url.scheme(),
                    data_slice=url.data_slice(),
                )
                sub_configuration.output_volume.url = sub_part_url
            output_obj = sub_configuration.output_volume
        else:
            raise TypeError(f"{type(sub_configuration)} not handled")

        working_directory = get_working_directory(output_obj)
        if working_directory is not None:
            script_dir = working_directory
        else:
            script_dir = "./"

        # save sub part nabu configuration file
        slurm_script_name = f"{stitch_prefix}_part_{i_sub_part}.sh"
        nabu_script_name = f"{stitch_prefix}_part_{i_sub_part}.conf"

        os.makedirs(script_dir, exist_ok=True)
        default_config = get_default_stitching_config(stitching_type)
        default_config.pop(SLURM_SECTION, None)
        generate_nabu_configfile(
            fname=os.path.join(script_dir, nabu_script_name),
            default_config=default_config,
            comments=True,
            sections_comments=_SECTIONS_COMMENTS,
            prefilled_values=sub_configuration.to_dict(),
            options_level="advanced",
        )

        command = f"python3 -m nabu.app.stitching {os.path.join(script_dir, nabu_script_name)}"
        script = (command,)

        job = SBatchScriptJob(
            slurm_config=slurm_configuration,
            script=script,
            script_path=os.path.join(script_dir, slurm_script_name),
            clean_script=clean_script,
            working_directory=working_directory,
        )
        job.overwrite = True
        if yield_configuration:
            yield job, sub_configuration
        else:
            yield job


def split_slices(slices: Union[slice, tuple], n_parts: int):
    if not isinstance(n_parts, int):
        raise TypeError(f"n_parts should be an int. {type(n_parts)} provided")
    if isinstance(slices, slice):
        assert isinstance(slices.start, int), "slices.start must be an integer"
        assert isinstance(slices.stop, int), "slices.stop must be an integer"
        start = stop = slices.start
        steps_size = int(numpy.ceil((slices.stop - slices.start) / n_parts))
        while stop < slices.stop:
            stop = min(start + steps_size, slices.stop)
            yield slice(start, stop, slices.step)
            start = stop
    elif isinstance(slices, (tuple, list)):
        start = stop = 0
        steps_size = int(numpy.ceil(len(slices) / n_parts))
        while stop < len(slices):
            stop = min(start + steps_size, len(slices))
            yield slices[start:stop]
            start = stop
    else:
        raise TypeError(f"slices type ({type(slices)}) is not handled. Must be a slice or an Iterable")


def get_working_directory(obj: TomoObject) -> Optional[str]:
    """
    return working directory for a specific TomoObject
    """
    if not isinstance(obj, TomoObject):
        raise TypeError(f"obj should be an instance of {TomoObject}. {type(obj)} provided")
    if isinstance(obj, (HDF5Volume, MultiTIFFVolume)):
        if obj.file_path is None:
            return None
        else:
            return os.path.abspath(os.path.dirname(obj.file_path))
    elif isinstance(obj, VolumeSingleFrameBase):
        if obj.data_url is not None:
            return os.path.abspath(obj.data_url.file_path())
        else:
            return None
    elif isinstance(obj, EDFTomoScan):
        return obj.path
    elif isinstance(obj, HDF5TomoScan):
        if obj.master_file is None:
            return None
        else:
            return os.path.abspath(os.path.dirname(obj.master_file))
    else:
        raise RuntimeError(f"obj type not handled ({type(obj)})")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/05/2022"


import numpy
from typing import Optional
from silx.utils.enum import Enum as _Enum
from nabu.misc import fourier_filters
from scipy.fft import rfftn as local_fftn
from scipy.fft import irfftn as local_ifftn


class OverlapStitchingStrategy(_Enum):
    MEAN = "mean"
    COSINUS_WEIGHTS = "cosinus weights"
    LINEAR_WEIGHTS = "linear weights"
    CLOSEST = "closest"
    IMAGE_MINIMUM_DIVERGENCE = "image minimum divergence"
    HIGHER_SIGNAL = "higher signal"


DEFAULT_OVERLAP_STRATEGY = OverlapStitchingStrategy.COSINUS_WEIGHTS

DEFAULT_OVERLAP_SIZE = None  # could also be an int
# default overlap to be take for stitching. Ig None: take the largest possible area


class OverlapKernelBase:
    pass


class ZStichOverlapKernel(OverlapKernelBase):
    """
    class used to define overlap between two scans and create stitch between frames (`stitch` function)
    """

    DEFAULT_HIGH_FREQUENCY_THRESHOLD = 2

    def __init__(
        self,
        frame_width: int,
        stitching_strategy: OverlapStitchingStrategy = DEFAULT_OVERLAP_STRATEGY,
        overlap_size: int = DEFAULT_OVERLAP_SIZE,
        extra_params: Optional[dict] = None,
    ) -> None:
        """ """
        from nabu.stitching.config import KEY_THRESHOLD_FREQUENCY  # avoid acylic import

        if not isinstance(overlap_size, int) and overlap_size > 0:
            raise TypeError(
                f"overlap_size is expected to be a positive int, {overlap_size} - not {overlap_size} ({type(overlap_size)})"
            )
        if not isinstance(frame_width, int) or not frame_width > 0:
            raise TypeError(
                f"frame_width is expected to be a positive int, {frame_width} - not {frame_width} ({type(frame_width)})"
            )

        self._overlap_size = abs(overlap_size)
        self._frame_width = frame_width
        self._stitching_strategy = OverlapStitchingStrategy.from_value(stitching_strategy)
        self._weights_img_1 = None
        self._weights_img_2 = None
        if extra_params is None:
            extra_params = {}
        self._high_frequency_threshold = extra_params.get(
            KEY_THRESHOLD_FREQUENCY, self.DEFAULT_HIGH_FREQUENCY_THRESHOLD
        )

    def __str__(self) -> str:
        return f"z-stitching kernel (policy={self.stitching_strategy.value}, overlap_size={self.overlap_size}, frame={self._frame_width})"

    @staticmethod
    def __check_img(img, name):
        if not isinstance(img, numpy.ndarray) and img.ndim == 2:
            raise ValueError(f"{name} is expected to be 2D numpy array")

    @property
    def overlap_size(self) -> int:
        return self._overlap_size

    @overlap_size.setter
    def overlap_size(self, size: int):
        if not isinstance(size, int):
            raise TypeError(f"height expects a int ({type(size)} provided instead)")
        if not size >= 0:
            raise ValueError(f"height is expected to be positive")
        self._overlap_size = abs(size)
        # update weights if needed
        if self._weights_img_1 is not None or self._weights_img_2 is not None:
            self.compute_weights()

    @property
    def img_2(self) -> numpy.ndarray:
        return self._img_2

    @property
    def weights_img_1(self) -> Optional[numpy.ndarray]:
        return self._weights_img_1

    @property
    def weights_img_2(self) -> Optional[numpy.ndarray]:
        return self._weights_img_2

    @property
    def stitching_strategy(self) -> OverlapStitchingStrategy:
        return self._stitching_strategy

    def compute_weights(self):
        if self.stitching_strategy is OverlapStitchingStrategy.MEAN:
            weights_img_1 = numpy.ones(self._overlap_size) * 0.5
            weights_img_2 = weights_img_1[::-1]
        elif self.stitching_strategy is OverlapStitchingStrategy.CLOSEST:
            n_item = self._overlap_size // 2 + self._overlap_size % 2
            weights_img_1 = numpy.concatenate(
                [
                    numpy.ones(n_item),
                    numpy.zeros(self._overlap_size - n_item),
                ]
            )
            weights_img_2 = numpy.concatenate(
                [
                    numpy.zeros(n_item),
                    numpy.ones(self._overlap_size - n_item),
                ]
            )
        elif self.stitching_strategy is OverlapStitchingStrategy.LINEAR_WEIGHTS:
            weights_img_1 = numpy.linspace(1.0, 0.0, self._overlap_size)
            weights_img_2 = weights_img_1[::-1]
        elif self.stitching_strategy is OverlapStitchingStrategy.COSINUS_WEIGHTS:
            angles = numpy.linspace(0.0, numpy.pi / 2.0, self._overlap_size)
            weights_img_1 = numpy.cos(angles) ** 2
            weights_img_2 = numpy.sin(angles) ** 2
        elif self.stitching_strategy in (
            OverlapStitchingStrategy.IMAGE_MINIMUM_DIVERGENCE,
            OverlapStitchingStrategy.HIGHER_SIGNAL,
        ):
            # those strategies are not using constant weights but have treatments depending on the provided img_1 and mg_2 during stitching
            return
        else:
            raise NotImplementedError(f"{self.stitching_strategy} not implemented")

        self._weights_img_1 = weights_img_1.reshape(-1, 1) * numpy.ones(self._frame_width).reshape(1, -1)
        self._weights_img_2 = weights_img_2.reshape(-1, 1) * numpy.ones(self._frame_width).reshape(1, -1)

    def stitch(self, img_1, img_2, check_input=True) -> tuple:
        """Compute overlap region from the defined strategy"""
        if check_input:
            self.__check_img(img_1, "img_1")
            self.__check_img(img_2, "img_2")

            if img_1.shape != img_2.shape:
                raise ValueError(
                    f"both images are expected to be of the same shape to apply stitch ({img_1.shape} vs {img_2.shape})"
                )

        if self._stitching_strategy is OverlapStitchingStrategy.IMAGE_MINIMUM_DIVERGENCE:
            return (
                compute_image_minimum_divergence(
                    img_1=img_1, img_2=img_2, high_frequency_threshold=self._high_frequency_threshold
                ),
                None,
                None,
            )
        elif self._stitching_strategy is OverlapStitchingStrategy.HIGHER_SIGNAL:
            return (
                compute_image_higher_signal(img_1=img_1, img_2=img_2),
                None,
                None,
            )
        else:
            if self.weights_img_1 is None or self.weights_img_2 is None:
                self.compute_weights()

            return (
                img_1 * self.weights_img_1 + img_2 * self.weights_img_2,
                self.weights_img_1,
                self.weights_img_2,
            )


def compute_image_minimum_divergence(img_1: numpy.ndarray, img_2: numpy.ndarray, high_frequency_threshold):
    """
    Algorithm to improve treatment of high frequency.

    It split the two images into two parts: high frequency and low frequency.

    The two low frequency part will be stitched using a 'sinusoidal' / cosinus weights approach.
    When the two high frequency part will be stitched by taking the lower divergent pixels
    """

    # split low and high frequencies
    def split_image(image: numpy.ndarray, threshold: int) -> tuple:
        """split an image to return (low_frequency, high_frequency)"""

        lowpass_filter = fourier_filters.get_lowpass_filter(
            image.shape[-2:],
            cutoff_par=threshold,
            use_rfft=True,
            data_type=image.dtype,
        )
        highpass_filter = fourier_filters.get_highpass_filter(
            image.shape[-2:],
            cutoff_par=threshold,
            use_rfft=True,
            data_type=image.dtype,
        )
        low_fre_part = local_ifftn(local_fftn(image, axes=(-2, -1)) * lowpass_filter, axes=(-2, -1)).real
        high_fre_part = local_ifftn(local_fftn(image, axes=(-2, -1)) * highpass_filter, axes=(-2, -1)).real
        return (low_fre_part, high_fre_part)

    low_freq_img_1, high_freq_img_1 = split_image(img_1, threshold=high_frequency_threshold)
    low_freq_img_2, high_freq_img_2 = split_image(img_2, threshold=high_frequency_threshold)

    # handle low frequency
    low_freq_stitching_kernel = ZStichOverlapKernel(
        frame_width=img_1.shape[1],
        stitching_strategy=OverlapStitchingStrategy.COSINUS_WEIGHTS,
        overlap_size=img_1.shape[0],
    )
    low_freq_stitched = low_freq_stitching_kernel.stitch(
        img_1=low_freq_img_1,
        img_2=low_freq_img_2,
        check_input=False,
    )[0]

    # handle high frequency
    mean_high_frequency = numpy.mean([high_freq_img_1, high_freq_img_2])
    assert numpy.isscalar(mean_high_frequency)
    high_freq_distance_img_1 = numpy.abs(high_freq_img_1 - mean_high_frequency)
    high_freq_distance_img_2 = numpy.abs(high_freq_img_2 - mean_high_frequency)
    high_freq_stitched = numpy.where(
        high_freq_distance_img_1 >= high_freq_distance_img_2, high_freq_distance_img_2, high_freq_distance_img_1
    )

    # merge back low and high frequencies together
    def merge_images(low_freq: numpy.ndarray, high_freq: numpy.ndarray) -> numpy.ndarray:
        """merge two part of an image. The low frequency part with the high frequency part"""
        return low_freq + high_freq

    return merge_images(low_freq_stitched, high_freq_stitched)


def compute_image_higher_signal(img_1: numpy.ndarray, img_2: numpy.ndarray):
    """
    the higher signal will pick pixel on the image have the higher signal.
    A use case is that if there is some artefacts on images which creates stripes (from scintillator artefacts for example)
    it could be removed from this method
    """
    # note: to be think about. But maybe it can be interesting to rescale img_1 and img_2
    # to ge something more coherent
    return numpy.where(img_1 >= img_2, img_1, img_2)



import numpy as np
from ..misc.processing_base import ProcessingBase
from .utils import get_opencl_context, __has_pyopencl__

if __has_pyopencl__:
    import pyopencl as cl
    import pyopencl.array as parray


class OpenCLProcessing(ProcessingBase):
    _array_class = parray.Array

    def __init__(self, ctx=None, device_type="all", queue=None, **kwargs):
        """
        Initialie a OpenCLProcessing instance.

        Parameters
        ----------
        ctx: pycuda.driver.Context, optional
            Existing context to use. If provided, do not create a new context.
        cleanup_at_exit: bool, optional
            Whether to clean-up the context at exit.
            Ignored if ctx is not None.
        """
        super().__init__()
        if queue is not None:
            # re-use an existing queue. In this case the this instance is mostly for convenience
            ctx = queue.context
        if ctx is None:
            self.ctx = get_opencl_context(device_type=device_type, **kwargs)
        else:
            self.ctx = ctx
        if queue is None:
            queue = cl.CommandQueue(self.ctx)
        self.queue = queue

    # TODO push_context, pop_context ?

    def _allocate_array_mem(self, shape, dtype):
        return parray.zeros(self.queue, shape, dtype)
